import glob
import os
import pprint
import textwrap

func_names = []
func_codes = []

CONVERTER = {
    "": lambda v, n, id: v or 0,
    "string": lambda v, n, id: str(v) if v else "",
    "int": lambda v, n, id: int(str(v or 0).split(".")[0]),
    "float": lambda v, n, id: float(str(v or 0)),
    "bool": lambda v, n, id: v != "FALSE",
    "array": lambda v, n, id: eval(f'[{v.replace("|",",")}]') if v else [],
    "array_str": lambda v, n, id: ["%s" % e for e in v.split("|")]
    if v else [],
    "array_bool": lambda v, n, id: [e != "FALSE" for e in v.split("|")]
    if v
    else [],
    "dict": lambda v, n, id: eval(f'{{{v.replace("|",",")}}}')
    if v else {},
    "function": lambda v, n, id: make_func(v or "pass", n, id),
}


def make_func(v, n, id):
    func_name = f"{n}_{id}"

    func_code = textwrap.dedent(
        f"""
static func {func_name}(args=[]):
{textwrap.indent(v, "    ")}
"""
    )
    func_names.append(func_name)
    func_codes.append(func_code)
    return func_name


def generator(data: dict, CFG: dict):

    # 表格数据脚本模板
    template = """
    extends Reference
    class Function extends Reference:

        var func_name
        var script_path

        func _init(script_path, func_name):
            self.func_name = func_name
            self.script_path = script_path

        func call(args=[]):
            var this_script = load(script_path)
            return this_script.call(self.func_name, args)

    var None = null
    var False = false
    var True = true

    var data = \\
    {data}

    {funcs}
    """
    template = textwrap.dedent(template)

    # 在这里面写你的加工逻辑
    field_names = data["define"]["name"]
    field_types = data["define"]["type"]
    table = {}
    for row in data["table"]:

        id_type = field_types[0]
        id_cvt = CONVERTER[id_type] if id_type in CONVERTER else CONVERTER[""]
        id = id_cvt(row[0], None, None)

        row_data = {}
        for index, value in enumerate(row):
            field_name: str = field_names[index]
            field_type = field_types[index]
            if field_name.startswith(CFG["settings"]["ignore_field_mark"]):
                continue

            if field_type in CONVERTER:
                cvt = CONVERTER[field_type]
            else:
                cvt = CONVERTER[""]

            row_data[field_name] = cvt(value, field_name, id)
        table[row_data["id"]] = row_data  # 我们规定第一个字段是ID字段

    funcs = textwrap.dedent("\n".join(func_codes))

    code = template.format(
        data=pprint.pformat(table, indent=2, width=10000000), funcs=funcs
    )

    code = textwrap.dedent(code)

    for func_name in func_names:
        code = code.replace(
            f"'{func_name}'",
            f"Function.new(self.get_script().get_path(),'{func_name}')"
        )

    return code, "gd"


def completed(CFG: dict):

    output_path = CFG["settings"]["output"]
    settings_file_path = os.path.join(output_path, "settings.gd")
    project_root = CFG["settings"]["project_root"]

    lines = []

    for path in glob.glob(f"{output_path}/**/*.*", recursive=True):
        if path == settings_file_path:
            continue  # 跳过 settings.gd
        basename = os.path.basename(path)
        setting_name = os.path.splitext(basename)[0]
        relpath = os.path.relpath(path, project_root).replace("\\", "/")
        lines.append(f"var {setting_name} = load('res://{relpath}').new()")

    # 去掉缩进
    code = textwrap.dedent(
        """
    extends Node
    # 这个脚本你需要挂到游戏的Autoload才能全局读表

    {refs_code}
    """
    )
    refs_code = "\n".join(lines)

    code = code.format(refs_code=refs_code)

    return code



import importlib
import os
import xlwings as xw
import glob


from .logger import logger
from toml import load, dump
from .generator.godot_script import (
    generator as GodotScriptGenerator,
    completed as GodotScriptCompleted
)

CFG = {
    "settings": {
        "ignore_sheet_mark": "~",  # 带星号的sheet不会导出
        "ignore_field_mark": "*",  # 字段忽略标志
        "custom_generator": "",  # 自定义加工脚本
        "completed_hook": "",  # 导表结束后调用脚本
        "input": "data",  # 导入目录
        "output": "dist",  # 导出目录
        # 导出到项目的路径，比如你的 output="../Godot项目/Scripts/Settings"，
        # 那么project_root="../Godot项目"
        "project_root": "../",
    }
}


def load_toml_config(filename: str = "export.toml"):
    """加载toml配置文件

    Raises:
        FileNotFoundError: 配置文件不存在的时候抛出

    Returns:
        dict: 配置数据字典，字段定义看CFG变量
    """

    local_cfg_path = os.path.abspath(filename)
    if os.path.exists(local_cfg_path):
        cfg_data = load(local_cfg_path)
        CFG["settings"].update(cfg_data["settings"])
    else:
        logger.warning(f"配置文件 {local_cfg_path} 不存在！将使用默认配置尝试导出。")


def create_default_toml_config():
    """
    用CFG模板创建一个默认的toml配置文件
    """
    with open("export.toml", "w+", newline="\n", encoding="utf-8") as f:
        dump(CFG, f)


def on_completed():
    """
    导表结束钩子
    """
    completed_hook = CFG["settings"]["completed_hook"]

    completed = GodotScriptCompleted

    if completed_hook:

        if os.path.isfile(completed_hook):
            logger.info(f"尝试加载用户自定义结束钩子 {completed_hook}")
            with open(completed_hook) as f:
                code = f.read()
                exec(code)
        else:
            try:
                logger.info(f"尝试加载内置结束钩子 {completed_hook}")
                module = importlib.import_module(completed_hook)
                completed = getattr(module, 'completed', completed)
            except Exception:
                logger.warning(f"{completed_hook} 不是内置结束钩子！请检查配置。")
                pass

    code = completed(CFG)
    code = "# 本文件由代码生成，不要手动修改" + code
    output_path = CFG["settings"]["output"]

    settings_file_path = os.path.join(output_path, "settings.gd")
    with open(settings_file_path, "w", newline="\n", encoding="utf-8") as f:
        f.write(code)
    logger.info("导表结束：Setting.gd 更新完毕!")


def excel2dict(workbook: xw.Book):
    """Excel表转字典对象

    Args:
        workbook (xw.Book): 工作簿
    """
    abspath: str = workbook.fullname
    abssource = os.path.abspath(CFG["settings"]["input"])
    absoutput = os.path.abspath(CFG["settings"]["output"])
    ignore_sheet_mark = CFG["settings"]["ignore_sheet_mark"]
    abspath_no_ext = os.path.splitext(abspath)[0]

    if not abspath.startswith(abssource):
        logger.error(f"{abspath} 不是配置表目录下的配置")
        return

    # 过滤出不带*的表名
    sheets = filter(
        lambda sheet: not sheet.name.startswith(ignore_sheet_mark),
        workbook.sheets
    )
    generator = GodotScriptGenerator  # 默认导出器

    custom_generator = CFG["settings"]["custom_generator"]
    if custom_generator:
        if os.path.isfile(custom_generator):
            with open(custom_generator) as f:
                custom_code = f.read()
                exec(custom_code)  # 执行后generator会被替换成字自定义generator
                logger.info(f"使用 {custom_generator} 自定义导出器")
        else:
            try:
                logger.info(f"尝试加载内置导出器 {custom_generator}")
                module = importlib.import_module(custom_generator)
                generator = getattr(module, 'completed', generator)
            except Exception:
                logger.warning(f"{custom_generator} 不是内置导出器！请检查配置。")
                pass

    for sheet in sheets:
        try:
            # 解析表名
            if "-" in sheet.name:
                org_name, rename = sheet.name.split("-")
            else:
                org_name, rename = sheet.name, None

            data = {}
            row_values = sheet.range("A1").expand().raw_value
            data["define"] = {
                "type": row_values[0],
                "desc": row_values[1],
                "name": row_values[2],
            }
            data["table"] = list(row_values[3:])

            relative_path = os.path.join(
                abspath_no_ext.replace(abssource, ""), rename or org_name
            )
            output = absoutput + relative_path

            output_dirname = os.path.dirname(output)
            if not os.path.exists(output_dirname):
                os.makedirs(output_dirname)

            code, ext = generator(data, CFG)  # 将字典传递给导出器
            code = "# 本文件由代码生成，不要手动修改\n"+code

            output = f"{output}.{ext}"
            with open(output, 'w', encoding='utf-8', newline='\n') as f:
                f.write(code)
                logger.info(f"导出: {abspath}:{sheet.name} => {output}")
        except Exception:
            logger.error(
                f"{sheet.name} 导出失败！"
            )

    workbook.close()


def gen_all(cwd: str = "."):
    os.chdir(cwd)
    load_toml_config()

    abssource = os.path.abspath(CFG["settings"]["input"])
    with xw.App(visible=False) as app:
        exts = [".xlsx", ".xls"]
        for ext in exts:
            full_paths = glob.glob(f"{abssource}/**/*{ext}", recursive=True)
            for full_path in full_paths:
                filename = os.path.basename(full_path)
                if filename.startswith("~$"):
                    logger.warning(f"{filename} 不是配置表，跳过！")
                    continue
                book = app.books.open(full_path)
                try:
                    excel2dict(book)
                except Exception:
                    logger.error(
                        f"{book.fullname}导出失败！"
                    )
        on_completed()


def gen_one(path, cwd: str = "."):
    with xw.App(visible=False) as app:
        book = app.books.open(path)
        try:
            excel2dict(book)
        except Exception:
            logger.error(
                f"{book.fullname}导出失败！"
            )
        on_completed()



import shutil
import click
import textwrap
import os
import pkg_resources
import excelexporter.exporter as exporter


@click.group()
def main():
    """
    ===============================
    Excel表导出工具
    ===============================
    """
    pass


@main.command()
@click.argument(
    "filename", nargs=1,
    type=click.Path(dir_okay=True),
    default="custom_generator"
)
def create_generator(filename="custom_generator.py"):
    """
    创建自定义生成器脚本
    """
    code = """
    # 传入到此脚本的变量
    # data : 读取sheet表加工成的字典
    # {
    #   "define":{
    #       "types": ['int'...],  -- 读取的表中第一行类型定义
    #           "desc": ['描述'...], -- 读取的第二行字段描述
    #           "name": ['id'...] --读取的第三行字段名
    #   },
    #   "table":[...] -- 每一行数据的值
    # }
    # data只是把表转成字典，你要怎么加工这个字典生成你想要数据格式并生成文件需要
    # 你自己在这个脚本里实现
    #
    # output: 导出文件。
    # 这个路径只有文件名，没有扩展名，例子： D://dist/道具/item
    # 如果你需要有扩展名，自己加 output+=".gd"，然后自己写入文件。

    # CFG 就是export.toml配置字典，你在哪个目录下执行ee命令，那么CFG就读哪个目录的export.toml
    # CFG["settings"]["input"] -- 读取配置表目录路径
    # 内置jinja包，你可以：
    # import jinja
    # 然后使用jinja模板来生成你的文件

    def generator(data:dict, CFG:dict) -> (str, str):
        # 在这里面写你的加工逻辑
        # data是excel sheet转成字典
        # CFG字典内容跟export.toml是一致的，你可以通过CFG获取到input、output、project_root
        # 返回值 （你生成的代码, 导出保存的后缀名（例如gd、tres））

        code = ""
        return code, 'gd'

    """
    with open(filename + ".py", "w", encoding="utf-8") as f:
        f.write(textwrap.dedent(code))


@main.command()
def create_completed_hook():
    """
    创建自定义导表完成钩子脚本
    """
    code = """
    # 钩子脚本组要是用于在导表结束后处理别的事务用
    # gen-all命令会在所有表导完后调用
    # gen-one命令会在导完这个表后调用

    # 最典型的用法是用来搜集所有配置表生成读表util:

    # class_name Settiings
    # settings.gd
    # const var Item = preload("res://Data/Item/Item.gd")
    # ...

    # 以后在代码里想要读这个表你可以
    # Settings.Item.data()

    # todo:具体要怎么实现看自己需求

    def completed(CFG:dict):
        # CFG字典内容跟export.toml是一致的，你可以通过CFG获取到input、output、project_root
        # 返回值 你生成的代码
        # todo: 你要在导表结束后做什么
        
        code = ""
        return code

    """
    code = textwrap.dedent(code)
    with open("completed_hook.py", "w", encoding='utf-8') as f:
        f.write(code)


@main.command()
def init():
    """
    生成默认配置表项目
    """
    exporter.create_default_toml_config()
    input_dir = exporter.CFG["settings"]["input"]
    output_dir = exporter.CFG["settings"]["output"]

    template_xlsx_path = pkg_resources.resource_filename(
        __package__,
        "template/示例.xlsx"
    )

    os.mkdir(input_dir)
    os.mkdir(output_dir)

    shutil.copy(template_xlsx_path, input_dir)


@main.command()
@click.option("--cwd", default=".", help="工作目录，执行命令所在的目录")
def gen_all(cwd):
    """
    导出所有表
    """
    exporter.gen_all(cwd)


@main.command()
@click.option("--cwd", default=".", help="工作目录，执行命令所在的目录")
@click.argument("file", type=click.Path(True))
def gen_one(file: str, cwd):
    """
    打开并导出整张excel表
    """
    exporter.gen_one(file, cwd)


if __name__ == "__main__":
    main()
